/************************************************************************/
/*                                                                      */
/*   Program Name :  octad2vdif                                         */
/*                                                                      */
/*     Octad2vdif is OCTAD raw data to VDIF file format converter       */
/*                                                                      */
/*     Enviroment variables                                             */
/*           VDIFDIR   ---   default directory for VDIF data            */
/*           OCTADDIR  ---   default directory for OCTAD raw data       */
/*                                                                      */
/*            Ver 1.10  2014-05-27                                      */
/*                                                                      */
/*      Copyright (c) 2014-   T.Kondo/NICT All Right Reserved           */
/*                                                                      */
/************************************************************************/
/*
     2014-05-18  -- Ver 1.00  start development based on "k5tom5b"
	 2014-05-21  -- Ver 1.01
	 2014-05-27  -- Ver 1.10


   how to run
          octad2vdif jjj|yyyymmdd|yyyyddd hh:mm:ss
                        octad_name [vdif_name] [options]
          

    
    where    
             jjj|yyyymmdd|yyyyddd  -- 
                     jjj      : lower 3digits of MJD 
                       or
                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month
                       or
                     yyyyddd  : 4 digit year, 3 digit day of year

                        of observation date
			 hh:mm:ss ---  UTC of 1st data
             vdif_name  -- VDIF file name to be created

             options (any order)
			   -ad adbit ------ set AD resolution to adbit bits (default is 1)
			   -fs[mhz] fsampl -- set sampling frequency to fsampl MHz (default 2048)
			   -nch numch   --- set # of channels to numch (default 1)
               -o vdif_name  -- VDIF file name to be created
                                 (default : see below)
               -d vdif_dir   -- directory for VDIF file out
                                 (default  : input file directory)
               -s strat_sec  -- offset time from data head (sec)
                                 (default :  0 )
               -p period     -- data period to convert (sec)
                                 (default : all)
			   -id sid       -- Station ID (A2) (default blank)
			   -das das      -- Station DAS (A8) (default " kashima")
			   -fr[amelength] frame_bytes -- Set frame length without header in byte
			                          (default: 1280)
			   -num[frame_sec] nframe -- Set # of frames per second
			                          (default: 2048e6/(1280*8) = 200000)

               -date jjj|yyyymmdd|yyyyddd  -- 
                     jjj      : lower 3digits of MJD 
                       or
                     yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month
                       or
                     yyyyddd  : 4 digit year, 3 digit day of year 

  VDIF naming rule
           If VDIF name is not given,  OCTAD name of which extension is replaced by "vdif"
		   will be VDIF name. If OCTAD name doesnt have extension, then ".vdif" will be
		   added.
		   If vdif directory is set, VDIF file destinated to VDIF directory.
		   If VDIF name is set and acompanied with full path, VDIF name doesn't change.

	   Examples
	       octad2vdif 2014/05/22 12:30:40 /home/vlbi/abcd12345.raw
		         VDIF name will be /home/vlbi/abcd12345.raw

	       octad2vdif 2014/05/22 12:30:40 /home/vlbi/abcd12345.raw -d /home/vdif/work
		         VDIF name will be /home/vdif/work/abcd12345.raw

	       octad2vdif 2014/05/22 12:30:40 /home/vlbi/abcd12345.raw test.vdif
		         VDIF name will be test.vdif

	       octad2vdif 2014/05/22 12:30:40 /home/vlbi/abcd12345.raw test.vdif -d /home/test
		         VDIF name will be /home/test/test.vdif
		    
		   


  Enviroment variables
          OCTADDIR    ---   default path to OCTAD raw data
          VDIFDIR     ---   default path to VDIF file


  AD bit is limited to 2bit max. 

Note:   only # of channels 1 is supported at the current version !!


*** sample run parameters ****
*/

#define PROGRAM_VERSION "octad2vdif (Ver 1.10  2014-05-27)"

//#define OLDSTYLE

//#include "libk5.h"      // common header for IP-VLBI utilities
//#include "libmk5tok5.h"  // library for Mark5 to K5 conversion
//#include "libvssp32.h"  // library for Mark5 to K5 conversion
#include "libvdif.h"


/* default conversion information file name set */
//#define INFONAME_DFLT "k5tom5binfo.txt"   // information file default name
/* environment variable set */
#define OCTADDIR "OCTADDIR" // the environment variable for OCTAD data default directory
/* environment variable for Mark5 vex file default directory set */
#define VDIFDIR "VDIFDIR" // the environment variable for VDIF file

/* default parameters  */
#define OCTAD_ADBIT  1     // AD resolution
#define OCTAD_SAMPLE 2048  // Sampling frequency in MHz
#define OCTAD_NUMCH  1     // Number of Channels
#define STATION_NAME "kashima" // station name at most 8 characters (A8)
#define EDV_NUMBER  1     // Extended format version number "1" for Kashima VDIF
#define FRAME_BYTES 1280  // Frame lenghth without header (in bytes)


typedef struct octad2vdifrunp     // run parameters for k5 to mark5 conversion program
{
    int ptype;   // run parameter type  =1  mode1,  =2 mode 2
    int channel;
    int group;
    int soffset;
    int period;
    //int reverse;
    //int makemode;
    //int scan;    // scan# for get MODE from VEX file (default=1) 
    //int odd;    // =1 4byte integer offset mode in case of 64 track
    int monit;   // =1  inofo_file monitor ON,  =0 no monitoring
    int yyyy, month, dd, ddd;  // observation date year month,day of month, day of year
	int hh,mm,ss; // UTC
    int mjd;   //  MJD
    int mjd3;   // lower 3 digit of MJD
    //int k5num;
    char vdifname[256]; 
    char octadname[256];
    char outdir[256];
    char sid[3];
	char das[9];
    char program_version[60]; // 2008-07-04

	int frame_bytes; // # of bytes in a frame (excepting header)
	int nframe_sec;  // # of frames per second

    // for mode2 only run parameres are as follows
    int numch;
    int adbit;
	int e2bit; // 2bit data encode mode
    int fmhz;
	int headlen;     // header length in 32-bit word (8)
} octad2vdifrunp_t;

void replace_extension(char *fname, char *extension)
{
/*
     replace file name extension by given extension
	 if extension portion is existed in the original name.
	 if it is not existed, extension is added as an extension
*/
	int i,n,len,len2;
	len=strlen(fname);
	len2=strlen(extension);
	n=len-1;
	while(n >=0){
		if(fname[n]=='.'){
			for(i=0; i<len2; i++){
				fname[n+i+1]=extension[i];
			}
			fname[n+len2+1]='\0';
			goto exit2return;
		}
		n--;
	}
	fname[len]='.'; fname[len+1]='\0';
	strcat(fname,extension);
exit2return:;
}

void combine_dir_name(char *fullpath, char *wdir, char *wname)
{
/*
     combine directory and name to get full path.
*/

	int i;
    strcpy(fullpath,wdir);
    i=strlen(fullpath);
    if(fullpath[i-1]!='\\' && fullpath[i-1]!='/'){
#ifdef WIN32
	   fullpath[i]='\\';
#else
	   fullpath[i]='/';
#endif
       fullpath[i+1]='\0';
	 }
	 strcat(fullpath,wname);
}


int hhmmss2sec(char *cbuf, int *hh, int *mm, int *ss)
{
    //  HH:MM:SS => seconds
    int n,len;
    int hh1,mm1,ss1,seconds;
    int p1,p2, n1st;
    char buf[10];
    len=strlen(cbuf);
    // check ':' position
    p1=0; p2=0;
    n1st=0;
    //printf("len %d\n",len);
    for(n=0; n<len; n++){
        if(cbuf[n]==':'){
            if(n1st==0){
                p1=n;
                n1st=1;
            } else {
                p2=n;
                n1st=2;
            }
        }
    }
    //printf("p1,p2 %d %d\n",p1,p2);
    if(n1st==0){ // no HH:MM:SS
       seconds=atoi(cbuf);
       hh1=seconds/3600;
       mm1= (seconds % 3600)/60;
       ss1= seconds % 60;
       *hh=hh1; *mm=mm1; *ss=ss1;
       return seconds;
    }

    strncpy(buf,cbuf,p1);
    buf[p1]='\0';
    hh1=atoi(buf);
    strncpy(buf,cbuf+p1+1,p2-p1-1);
    buf[p2-p1]='\0';
    mm1=atoi(buf);
    strncpy(buf,cbuf+p2+1,len-p2-1);
    buf[len-p2]='\0';
    ss1=atoi(buf);
    seconds=hh1*3600+mm1*60+ss1;
    *hh=hh1; *mm=mm1; *ss=ss1;
    //printf("HH MM SS %d %d %d\n",hh,mm,ss); // debug
    return seconds;
}



int leaf_year_check(int year)   // return 0 in case of leaf year
{
       /* leaf year check */
      if (year-(year/4)*4 == 0) 
      { 
          if (year-(year/100)*100 == 0) 
          { 
               if (year-(year/400)*400 == 0) { return 0;}
               return -1;
           }
           return 0;
       }
       return -1;
}

void ddd2md(int year, int ddd, int* month, int* mday)
{
      int md[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
      int i,td,wd ;
       /* leaf year check */
      if (leaf_year_check(year)==0)
      {
         md[1]=29;
      }
      *month=0;
      *mday=0;
      td=ddd;
      for (i=0; i< 12; i++)
      {
         wd=td-md[i];
         if (wd <=0)
         {
           *mday=td;
           *month=i+1;
           goto exithere;
          }
         td=wd;
       }
   exithere:;
 }

void md2ddd(int year, int* ddd, int month, int mday)
{
      int md[12] = {0,31,28,31,30,31,30,31,31,30,31,30};
      int i;
       /* leaf year check */
      if (leaf_year_check(year)==0)
      {
         md[2]=29;
      }
      *ddd=0;
      for (i=0; i< month; i++)
      {
        *ddd=*ddd+md[i];
      }
      *ddd += mday;
 }

/* UTC to MJD conversion */
int ut2mjd(double *d, int year, int month, int day,
           int hour, int min, int sec)
/*
  [in]
     year   -- year (yyyy)
     month  -- month (month)
     day    -- day of month (dd)
     hour   -- hour (hh)
     min    -- minute (mm)
     sec    -- second (ss)
  [return]
     ut2mjs -- MJD  (integer day)
     d      -- fractional part of MJD (day)
*/
{
    int  bc, gregory, mnth, y4; 
    //long j;
    int j;

	mnth=month;
	y4=year;
    bc = year <= 0;
    
    /* First day at Gregory's calendar */
    gregory = year > 1582 || (year == 1582 && month > 10) ||
              (year == 1582 && month == 10 && day >= 15);
            
    if (mnth <= 2) {
        y4--;
        mnth += 12;
    }
    
    if (hour < 12) {
        j = 0;
        *d = 0.5;
    } else {
        j = 1;
        *d = -0.5;
    }
    *d += (hour*3600L + min*60 + sec)/86400.0;
    
    j += bc ? (y4-3)/4 : y4/4;
    if (gregory) j+= 2 - y4/100 + y4/400;
    j += 1720994L + y4*365L + (mnth+1)*30 + (mnth+1)*3/5 + day;
    
	/* convert to MJD */
	if (hour < 12) {
		*d -= 0.5;
	} else {
		*d += 0.5;
		j -= 1L;
	}
    j -= 2400000L;
    return j;
}

/* MJD to UTC conversion */
void mjd2ut(int *year, int *month, int *day,
		   int *hour, int *min, int *sec, int j, double d)
/*
  [in]
     j      -- MJD  (integer day)
     d      -- fractional part of MJD (day)
  [return]
     year   -- year (yyyy)
     month  -- month (month)
     day    -- day of month (dd)
     hour   -- hour (hh)
     min    -- minute (mm)
     sec    -- second (ss)
*/
{
    int c, k, e, s;

    /* convert to JD */
    if( d >= 0.5) {
		d -= 0.5;
		j++;
	} else {
		d += 0.5;
	}
	j += 2400000L;

	/* JD to UTC */
    if (d >= 0.5) {
        j++;
        d -= 0.5;
    } else d += 0.5;

    if (j >= 2299161) j = j + 1 + (j-1867216)/36524 - (j-1867216)/146096;
    j += 1524;

    //c = (long)((j - 122.1)/365.25);
    c = (int)((j - 122.1)/365.25);
    k = 365L*c + c/4;
    //e = (long)((j - k)/30.6001);
    e = (int)((j - k)/30.6001);

    *year = c - 4716;
    *month = e - 1;
    if (*month > 12) {
       (*month) -= 12;
       (*year)++;
    }
    //*day = j - k - (long)(30.6*e);
    *day = j - k - (int)(30.6*e);

    //s = (long)(d * 86400 + 0.5);
    s = (int)(d * 86400 + 0.5);
    *hour = s / 3600;
    *min = (s % 3600) / 60;
    *sec = s % 60;
}


void param_get_octad2vdif(int argc, char *argv[],octad2vdifrunp_t *runp)                       
{
    int n,i;
    int len;
    int mode;
    char cbuf[256],buf[10];
    double dw; // work

    // initialize parameters

    /* common parameters */
    runp->ptype=1;  // run with information file
    runp->octadname[0]='\0';
    runp->vdifname[0]='\0';

    runp->soffset=0;
    runp->period=0;

    runp->outdir[0]='\0';
	runp->sid[0]='\0'; runp->sid[1]='\0'; runp->sid[2]='\0';
    runp->monit=0;

    runp->adbit =OCTAD_ADBIT;
    runp->fmhz  =OCTAD_SAMPLE;
	runp->numch =OCTAD_NUMCH;
    runp->e2bit=2;   // 2bit AD encode mode

	runp->frame_bytes=FRAME_BYTES;
	runp->nframe_sec=0;
	runp->headlen=8;  // fixed value

    runp->yyyy=0; runp->month=0; runp->dd=0; runp->ddd=0;
	runp->hh=0; runp->mm; runp->ss=0;
    runp->mjd=0; runp->mjd3=0;
	for(i=0; i<9; i++) runp->das[i]='\0';
	if(strlen(STATION_NAME) > 8){
		strncpy(runp->das,STATION_NAME,8);
	} else {
		strcpy(runp->das,STATION_NAME);
	}

    mode=0;
    /* 1st param check whether it is started with "-" or not  */
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        if(cbuf[0] == '-') mode=1;
    }

    if(mode == 1) {
        n=2;
        goto comm_anal;
    }

    /* 1st param should be either jjj(3digit MJD) or yyyymmdd or yyyyddd */
    if (argc>=2){
        strcpy(cbuf,*(argv+1));
        len=strlen(cbuf);
        //printf("len = %d\n",len);
        if(len==8){ // yyyymmdd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,2); buf[2]='\0';
			runp->month=atoi(buf);
            strncpy(buf,cbuf+6,2); buf[2]='\0';
            runp->dd=atoi(buf);
            //get MJD
            runp->mjd=ut2mjd(&dw,runp->yyyy,runp->month,runp->dd,0,0,0);
            //get total day 
            md2ddd(runp->yyyy,&runp->ddd,runp->month,runp->dd);
        } else if(len==7){ // yyyyddd
            strncpy(buf,cbuf,4); buf[4]='\0';
            runp->yyyy=atoi(buf);
            strncpy(buf,cbuf+4,3); buf[3]='\0';
            runp->ddd=atoi(buf);
            //get month day
            ddd2md(runp->yyyy,runp->ddd,&runp->month,&runp->dd);
            //get MJD
            runp->mjd=ut2mjd(&dw,runp->yyyy,runp->month,runp->dd,0,0,0);
        } else {
            runp->mjd = atoi(cbuf);
        }
        runp->mjd3 = runp->mjd % 1000;
    }

	// 2nd parameter should be HH:MM:SS
    if(argc >=3){
        strcpy(cbuf,*(argv + 2));
        i=hhmmss2sec(cbuf,&runp->hh,&runp->mm,&runp->ss);
    }

	// 3rd parameter 
    if(argc >=4){
        strcpy(cbuf,*(argv + 3));
        if(cbuf[0] == '-') goto comm_anal;  // go to command parameter get
        strcpy(runp->octadname,cbuf);
    }

	// 4th parameter

    if(argc >=5){
        strcpy(cbuf,*(argv + 4));
        if(cbuf[0] == '-') goto comm_anal;  // go to command parameter get
        strcpy(runp->vdifname,cbuf);
    }

comm_anal:

    /* command parameters get */
    if(cmnd_get(argc,argv,"-o",cbuf)==0) strcpy(runp->vdifname,cbuf);  // vdif name get
    if(cmnd_get_p(argc,argv,"-d",cbuf)==0) strcpy(runp->outdir,cbuf);  // out directory get (character length in command is also compared)
    if(cmnd_get(argc,argv,"-e",cbuf)==0) runp->e2bit = atoi(cbuf);      // 2bit encode select mode
    if(cmnd_get(argc,argv,"-s",cbuf)==0) runp->soffset = atoi(cbuf);      // start offset second
    if(cmnd_get(argc,argv,"-p",cbuf)==0) runp->period = atoi(cbuf);      // period
	if(cmnd_get(argc,argv,"-fs",cbuf)==0) runp->fmhz = atoi(cbuf);      // sampling frequency (MHz)
	if(cmnd_get(argc,argv,"-fr",cbuf)==0) runp->frame_bytes = atoi(cbuf);      // Frame length without header (in bytes)
	if(cmnd_get(argc,argv,"-num",cbuf)==0) runp->nframe_sec = atoi(cbuf);      // # of frames per second (in bytes)
	if(cmnd_get(argc,argv,"-id",cbuf)==0) strncpy(runp->sid,cbuf,2);  // out directory get
	if(cmnd_get_p(argc,argv,"-das",cbuf)==0){    //DAS name get
	    for(i=0; i<9; i++) runp->das[i]='\0';
		if(strlen(cbuf) > 8	){
			strncpy(runp->das,cbuf,8); 
		} else {

			strcpy(runp->das,cbuf);
		}
	}
	if(cmnd_get(argc,argv,"-ad",cbuf)==0) runp->adbit = atoi(cbuf);      // AD bit
}

void print_usage(void)
{
    printf("%s\n",PROGRAM_VERSION);
    printf("\n");
    printf("How to use\n");
    printf("\n");
    //printf("    k5tom5 yymmdd k51name [k52name k53name k54name] [options]\n");
	printf("    octad2vdif jjj|yyyymmdd|yyyyddd HH:MM:SS\n");
    printf("                   octad_name vdif_name [options]\n");
    printf("        or\n");
    printf("   where\n");
    printf("       jjj|yyyymmdd|yyyyddd  -- observation date in three ways as follows\n");
    printf("           jjj      : lower 3digits of MJD \n");
    printf("              or\n");
    printf("           yyyymmdd : 4 digit year, 2 digit month, 2 digit day of month\n");
    printf("              or\n");
    printf("           yyyyddd  : 4 digit year, 3 digit day of year\n");
	printf("       HH:MM:SS ---  UTC of start time\n");
    printf("       octad_name ---- OCTAD format file name\n");
    printf("       vdif_name  ---- VDIF file name to be created\n");
    printf("      options (any order)\n");
    printf("        -o vdif_name  -- VDIF name to be created\n");
    //printf("                          (default : see below)\n");
    printf("        -d vdif_dir   -- directory for VDIF file out\n");
    printf("                          (default  : OCTAD file directory)\n");
    printf("        -s strat_sec  -- offset time from data head (sec)\n");
    printf("                          (default :  0 )\n");
    printf("        -p period     -- data period to convert (sec)\n");
    printf("                          (default : all)\n");
    printf("        -e2bit n      -- 2bit data encode mode\n");
    printf("                          n=1 : offset binary 0 1 2 3\n");
    printf("                          n=2 : twisted offset binary 0 2 1 3\n");
    printf("                          n=3 : signed binary 1 0 2 3\n");
    printf("                          (default is 2)\n");
    printf("        -ad adbit     -- set AD resolution to adbit bits (default is 1)\n");
    printf("        -fs[mhz] fsampl -- set sampling frequency to fsampl MHz\n");
    printf("	                       (default 2048)\n");
    printf("        -fr[amelength] frame_bytes -- set frame length without header (in byte)\n");
    printf("	                       (default 1280)\n");
    printf("        -num[frame_sec] nframe -- Set # of frames per second\n");
    printf("                                 (default: 2048e6/(1280*8) = 200000)\n");
    printf("        -nch numch    -- set # of channels to numch (default 1)\n");
    //printf("        -monit        -- info_file monitor ON\n");
    printf("\n");
}

void monit_run_condition(octad2vdifrunp_t *runp, vdif_header_t *vhead)
// run parameter monitor
{
       printf(" ---------- RUN CONDITION ----------\n");
       printf("   yyyymmdd   : %04d%02d%02d\n",runp->yyyy,runp->month,runp->dd);
       printf("   yyyyddd    : %04d%03d\n",runp->yyyy,runp->ddd);
       printf("   jjj        : %03d\n",runp->mjd3);
       if(runp->yyyy != 0){
       printf("   MJD        : %d\n",runp->mjd);
       }
	   printf("   OCTD name  : %s\n",runp->octadname);

       printf("   VDIF name  : %s\n",runp->vdifname);
       printf("   outdir     : %s\n",runp->outdir);
       printf("   soffset    : %d\n",runp->soffset);
       printf("   period     : %d\n",runp->period);
       //printf("   e2bit      : %d\n",runp->e2bit);
	   printf("   AD bits    : %d\n",runp->adbit);
	   printf("   sampling   : %d MHz\n",runp->fmhz);
	   printf("   # of Chs   : %d\n",runp->numch);
	   printf("   station id (A2) : %s\n",runp->sid);
	   printf("   station das (A8): %s\n",runp->das);
	   printf("   Frame length (without header) : %d bytes\n",runp->frame_bytes);
	   printf("   Frame length (with header)    : %d bytes\n",runp->frame_bytes+runp->headlen*4);
	   printf("   # of frames per second        : %d\n",runp->nframe_sec);
	   printf("   start time : %4d/%02d/%02d %02d:%02d:%02d\n",
		   vhead->t.year,vhead->t.month,vhead->t.mday,vhead->t.hh,vhead->t.mm,vhead->t.ss);
	   printf("   start ref epo:  %d,   sec from epoch: %d\n", vhead->ref_epo,vhead->sec_fm_ref);
       //printf("   odd       : %d\n",runp->odd);
       //printf("   parity    : %d\n",runp->parity);
       //printf("   nrzm      : %d\n",runp->nrzm);
       //printf("   vlba      : %d\n",runp->vlba);
       //printf("   nbits     : %d\n",runp->nbits);
       printf(" -------------------------------------\n");

    printf("\n");
}

void ymdhms2vdiftime(timeval_vdif_t *t)
{
/*  
      convert normal time to VDIF time
	       T.Kondo   2014-05-18
		   
*/
	//int year0, month0;
	int day0=1;
	int mjd0;
	//double d0;
	int days, secs;
    timeval_vdif_t twork;

	//get ref epoch

	t->ref_epo=(t->year-2000)*2;
	if(t->month >=6) t->ref_epo++;

	// days from ref epoch 
	twork=*t;
	twork.mday=1;
	if(twork.month >=6){
		twork.month=6;
	} else {
		twork.month=1;
	}
    ut2mjd_v(&twork);
	mjd0=twork.mjd;

	twork.month=t->month;
	twork.mday=t->mday;
    ut2mjd_v(&twork);

	days=twork.mjd-mjd0;

	// seconds from epoch

	secs=days*86400+t->hh*3600+t->mm*60+t->ss;

	t->sec_fm_ref=secs;

}

void vdif_header_make_ini(u_int32_t *head, vdif_header_t *vhead)
{
/*
      VDIF header generation (VDIF Kashima format)
	  initial call
	  2014-05-18
*/

	head[0]=0x0;
	head[0]=(head[0] & 0x7FFFFFFF) | ((vhead->invalid & 0x1) << 31);
	head[0]=(head[0] & 0xBFFFFFFF) | ((vhead->legacy & 0x19) << 30);
	head[0]=(head[0] & 0xC0000000) | (vhead->sec_fm_ref & 0x3FFFFFFF);

	head[1]=0x0;
	head[1]=(head[1] & 0xC0FFFFFF) | ((vhead->ref_epo & 0x3F) << 24);
	head[1]=(head[1] & 0xFF000000) | (vhead->frame & 0xFFFFFF);

	head[2]=0x0;
	head[2]=(head[2] & 0x1FFFFFFF) | ((vhead->ver & 0x7) << 29);
	head[2]=(head[2] & 0xE0FFFFFF) | ((vhead->log2_chs & 0x1F) << 24);
	head[2]=(head[2] & 0xFF000000) | (vhead->frame_len64 & 0xFFFFFF);

	head[3]=0x0;
	head[3]=(head[3] & 0x7FFFFFFF) | ((vhead->d_type & 0x1) << 31);
	head[3]=(head[3] & 0x83FFFFFF) | ((vhead->bpsampl & 0x1F) << 26);
	head[3]=(head[3] & 0xFC00FFFF) | ((vhead->thread_id & 0x3FF) << 16);
	head[3]=(head[3] & 0xFFFF0000) | (vhead->station_id & 0xFF);

	head[4]=0x0;
	head[4]=(head[4] & 0x00FFFFFF) | ((vhead->edv & 0xFF) << 24);

	// VDIF KASHIMA
	head[4]=(head[4] & 0xFF7FFFFF) | ((vhead->uflag & 0x1F) << 23);
	head[4]=(head[4] & 0xFF800000) | (vhead->srate & 0x7FFFFF);

	head[5]= vhead->sync_block;
	strncpy((char *)head+24,vhead->das,8);

}

void vdif_header_make(u_int32_t *head, vdif_header_t *vhead)
{
/*
      VDIF header generation (VDIF Kashima format)
	  only possible updated part
	  2014-05-18
*/

	head[0]=(head[0] & 0x7FFFFFFF) | ((vhead->invalid & 0x1) << 31);
	head[0]=(head[0] & 0xC0000000) | (vhead->sec_fm_ref & 0x3FFFFFFF);

	head[1]=(head[1] & 0xC0FFFFFF) | ((vhead->ref_epo & 0x3F) << 24);
	head[1]=(head[1] & 0xFF000000) | (vhead->frame & 0xFFFFFF);

	//head[2]=(head[2] & 0x1FFFFFFF) | ((vhead->ver & 0x7) << 29);
	//head[2]=(head[2] & 0xE0FFFFFF) | ((vhead->log2_chs & 0x1F) << 24);
	//head[2]=(head[2] & 0xFF000000) | (vhead->frame_len64 & 0xFFFFFF);

	//head[3]=(head[3] & 0x7FFFFFFF) | ((vhead->d_type & 0x1) << 31);
	//head[3]=(head[3] & 0x83FFFFFF) | ((vhead->bpsampl & 0x1F) << 26);
	head[3]=(head[3] & 0xFC00FFFF) | ((vhead->thread_id & 0x3FF) << 16);
	//head[3]=(head[3] & 0xFFFF0000) | (vhead->station_id & 0xFF);

	//head[4]=(head[4] & 0x00FFFFFF) | ((vhead->edv & 0xFF) << 24);

	// VDIF KASHIMA
	//head[4]=(head[4] & 0xFF7FFFFF) | ((vhead->uflag & 0x1F) << 23);
	//head[4]=(head[4] & 0xFF800000) | (vhead->srate & 0x0x7FFFFF);

	//head[5]= vhead->sync_block;
	//strncpy((char *)head+24,vhead->das,8);

}

int main(int argc, char *argv[])
{
       int i;
       int iread,iwrite;
       u_int32_t *obuf;  // for 1 Frame data 2504 X 32 bits
       u_int32_t *head;   // for header 8 X 32 bits
       u_int32_t *idata;  // for data field 320 X 32 bits
       
       int nbuf;   //
       int nframe;

       int totch;  // total number of channels
       int bitsample;  // # of bits/sample

	   int period;

       int imax;
	   int wspersec,wsperframe,wmax,wres,wsuggested;
       //int bittable[64];
       int adjust_flag;   // flag to set conversion length integer second  1:set 0: no set
	   double fsample;
	   int sec_count; // 2012-12-03
       
       double telapsed;           // to measure processing time
	   char wname[100],wdir[256],odir[256];
	   int64_t f_size;
	   int64_t pos;
       time_t tstart, tfinish;    
       octad2vdifrunp_t runp;
       FILE *fi,*fo;
	   int flag_fi_open, flag_fo_open;
       vdif_header_t vhead;
       int sync_pat=VDIF_KASHIMA_SYNC;  // VDIF Kashima format sync pattern 0xACABFEED
       int imask[]={0,0x7FFFFFFF,0x3FFFFFFF,0,0xFFFFFFF,0,0,0,0xFFFFFF};
                 // masking table for 1,2,4,8 bit sampling data
       int *conv2bit;
       int conv2bit_1[]={0,1,2,3};   // offset binary (no conversion 2 bit)
       int conv2bit_2[]={0,2,1,3};   //  conversion 2 bit (default)
       int conv2bit_3[]={1,0,2,3};   //  conversion 2 bit singed integer

	   flag_fi_open=0; flag_fo_open=0;

       if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
          print_usage();
          exit(0);
       } 
       if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
          printf("\n");
          printf("%s\n",PROGRAM_VERSION);
          printf("\n");
          exit(0);
       } 
       
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for OCTAD data\n",OCTADDIR);
          printf("                      ( %s )\n",getenv(OCTADDIR));
          printf("        %s ---   default directory for VDIF data\n",VDIFDIR);
          printf("                      ( %s )\n",getenv(VDIFDIR));
          printf("\n");
          exit(0);
          }
       }

       strcpy(runp.program_version,PROGRAM_VERSION);

       /* run parameters get */
       param_get_octad2vdif(argc, argv, &runp); 

       if(runp.e2bit==1){
           conv2bit=conv2bit_1;
       } else if(runp.e2bit==3){
           conv2bit=conv2bit_3;
       } else {
           conv2bit=conv2bit_2;
       }

	   /* set up VDIF file name */

       parse_dir_and_name(runp.octadname,odir,wname);
	   if(runp.vdifname[0] == '\0'){  // VDIF name is OCTAD name (extension is replaced by "vdif"
		   strcpy(runp.vdifname,runp.octadname);
		   replace_extension(wname, "vdif");
		   if(runp.outdir[0] != '\0'){ // use given out directory
			   combine_dir_name(runp.vdifname, runp.outdir, wname);
		   } else {                    // use OCTAD directory
			   combine_dir_name(runp.vdifname, odir, wname);
		   }
	   } else {                            // VDIF name is given
           parse_dir_and_name(runp.vdifname,wdir,wname);
		   if( runp.outdir[0]!='\0'){               // use given out directory
			   combine_dir_name(runp.vdifname, runp.outdir, wname);
		   } else {
			   if(wdir[0]=='\0'){  // use OCTAD file directory
			       combine_dir_name(runp.vdifname, odir, wname);
			   }
		   }
	   }

	   //printf("Debug VDIF name : %s\n",runp.vdifname);


	   //nbuf=runp.frame_bytes/4 + runp.headlen;

       //obuf  = (u_int32_t *)malloc(nbuf*sizeof(u_int32_t)); 


       //setup_k5files_full_path(&runp);  // added on 2008-09-30


       adjust_flag=1;   // set period is integer second mode  2010-07-21

       /* Program Name Display */
       printf("************************************************************\n");
       printf("*    OCTAD to VDIF Data Format Converter                   *\n");
       printf("*      %s  by  T.KONDO/NICT *\n",runp.program_version);
       printf("************************************************************\n");
       printf("\n");



       tstart= time(NULL);



        /********************************/
        /* set internal-use parameters  */
        /********************************/


	   if(runp.adbit > 2){
           printf("This AD bits (%d) is not supported by the current version!\n",runp.adbit); 
           goto exitend;
       }

	   // set obuf length

	   fsample=(double)runp.fmhz*1.0e6;
	   wspersec=(int)fsample*runp.adbit;

	   if(runp.nframe_sec==0){ // check frame length
		   wsperframe=runp.frame_bytes*8;
		   wmax=wspersec/wsperframe;
		   wres=wspersec % wsperframe;
		   if(wres !=0){
			   printf(" # of samples per second (%d) devided by # of samples per frame (%d)\n",wspersec,wsperframe);
			   printf(" is not an integer!!\n");
			   wsuggested=wsperframe;
			   while((wspersec % wsuggested) !=0 ) {
				   runp.frame_bytes--;
				   wsuggested=runp.frame_bytes*8;
			   }
			   printf(" Advised frame length (without header) is %d bytes\n", runp.frame_bytes);
			   printf(" Try again with changing frame length parameter!\n");
			   goto exitend;
		   }
		   imax=wmax;
		   runp.nframe_sec=imax;
	   } else {  // calc # bytes per frame 
		   wres=(wspersec/32) % runp.nframe_sec;
		   if(wres !=0){
			   printf(" # of data (in 32-bit words) per second (%d)\n", wspersec/32);
			   printf("             devided by # of frames per second (%d)\n",runp.nframe_sec);
			   printf(" is not an integer!!\n");
			   wsuggested=runp.nframe_sec;
			   while(((wspersec/32) % wsuggested) !=0 ) {
				   runp.nframe_sec--;
				   wsuggested=runp.nframe_sec;
			   }
			   printf(" Advised # of frames per second is %d bytes\n", runp.nframe_sec);
			   printf(" Try again with changing # of frames per second!\n");
			   goto exitend;
		   }
		   runp.frame_bytes=((wspersec/32)/runp.nframe_sec)*4;
		   imax=runp.nframe_sec;
	   }


	   /* output buffer set */
	   nbuf=runp.frame_bytes/4 + runp.headlen;
       obuf  = (u_int32_t *)malloc(nbuf*sizeof(u_int32_t)); 

       head=obuf;
       idata=obuf+8;

	   vhead.frame_len64=nbuf/2;  // 2014-05-30

	   // set VDIF time

	   vhead.t.year=runp.yyyy;
	   vhead.t.month=runp.month;
	   vhead.t.mday=runp.dd;
	   vhead.t.hh=runp.hh;
	   vhead.t.mm=runp.mm;
	   vhead.t.ss=runp.ss;

	   // convert to VDIF time
	   ymdhms2vdiftime(&vhead.t);

	   // VDIF header fixed parameters
	   vhead.invalid=0;
	   vhead.legacy=0;
	   vhead.frame=0;
	   vhead.ver=0;
	   vhead.log2_chs=0;    // means 1ch
	   //vhead.frame_len64=nbuf/2 ;
	   vhead.d_type=0;
	   vhead.bpsampl=0;  // means 1 bit sampling
	   vhead.station_id=0;
	   //strcpy((char *)vhead.station_id,"kk");
	   vhead.edv=EDV_NUMBER;    // means VDIF Kashima Format
	   vhead.uflag=1;  // meand MHz
	   vhead.srate=runp.fmhz;
	   vhead.sync_block=sync_pat;
	   strcpy(vhead.das,runp.das);

	   // VDIF header variable parameters
	   vhead.thread_id=0;
	   vhead.ref_epo=vhead.t.ref_epo;
	   vhead.sec_fm_ref=vhead.t.sec_fm_ref;

       /* run condition monitor */

       monit_run_condition(&runp, &vhead);

       /* OCTAD data file open */

       totch=0;
       bitsample=0;
	   if((fi  = fopen( runp.octadname, "rb" )) == NULL ){
		   printf("Data File %s   Open Error! \n",runp.octadname);
           goto exitend;
       }
	   flag_fi_open=1;
	   f_size=get_file_size(runp.octadname);

       // start offset (just forwarding reading pointer by computing from sampling rate)
       if(runp.soffset > 0){
		   pos=(int64_t)runp.soffset*(int64_t)(wspersec/8);
		   if(pos > f_size){
			   printf(" %d sec forward is not possible (exceed data period)!\n",runp.soffset);
			   goto exitend;
		   } else {
		       fseek(fi,pos,SEEK_SET); // move pointer
               if( ferror(fi) ){
				   goto exitend;
			   }
			   printf(" %d sec forwarded.\n",runp.soffset);
		   }
	   }


       /* VDIF file create */

	   if( (fo  = fopen( runp.vdifname, "wb" )) == NULL ){
		   printf("  File  %s  Open Error! \n",runp.vdifname);
           goto exitend;
       }
	   flag_fo_open=1;
     

	   // initialize VDIF header
       vdif_header_make_ini(head, &vhead);

	   //printf("Debug %08X %08X %08X %08X\n",head[0],head[1],head[2],head[3]);
	   //printf("      %08X %08X %08X %08X\n",head[4],head[5],head[6],head[7]);

       /* conversion start message */

       printf("\n");
	   printf("OCTAD raw format file : %s\n",runp.octadname);
       printf("VDIF file (created) : %s\n",runp.vdifname);
//       printf("    Observation Date : 3digit MJD = %4d\n",runp.mjd3);
//       if(runp.yyyy != 0){
//       printf("                     :      Year  = %4d\n",runp.yyyy);
//       printf("                     :      Month = %4d\n",runp.mm);
//       printf("                     :      Day   = %4d\n",runp.dd);
//       printf("                     :  Total Day = %4d\n",runp.ddd);
//       }
//       if(adjust_flag ==1){
//       printf("    Conversion length (sec) :  %d\n",runp.period);
//       }
//       printf("\n");

//       hd.vlba=1;
       printf("\n");
       printf("                   TIME LABEL MONITOR\n");
       printf(" ==========================================================\n");   
       printf("        OCTAD-FILE                     VDIF-FILE\n");   
       printf("  YYYY/MM/DD HH:MM:SS             REF_EPOCH SEC_FROM_EPOCH\n");   
       printf(" ----------------------------------------------------------\n");   

       period=0;
	   sec_count=0;

       while(1){    // entry point of second loop

           nframe=0;  // frame # (in a second) counter reset
		   vhead.frame=0;

		   vhead.t.sec_fm_ref=vhead.sec_fm_ref;
           vdiftime2ymdhms(&vhead.t);   // get YYYY/MM/DD HH:MM:SS for debug
	       //printf("Debug %08X %08X %08X %08X\n",head[0],head[1],head[2],head[3]);  //debug
	       //printf("      %08X %08X %08X %08X\n",head[4],head[5],head[6],head[7]);  //debug
		   //print time label
		   printf("  %04d/%02d/%02d %02d:%02d:%02d",vhead.t.year,vhead.t.month,vhead.t.mday,
			   vhead.t.hh,vhead.t.mm,vhead.t.ss);
		   printf("               %4d    %10d\n",vhead.ref_epo,vhead.sec_fm_ref);
           for(i=0; i<imax; i++){  // 1 sec loop is imax'th Mark5 frame

               vdif_header_make(head,&vhead); 

               /* read OCTAD data corresponding to one frame of VDIF data */ 
			   iread = fread( idata, sizeof(int), runp.frame_bytes/4, fi);
               if (iread <= 0) goto exit_loop;  // exit loop  2008-07-04

               /* out VDIF frame data */

               iwrite=fwrite(obuf,sizeof(int),nbuf,fo);
               //printf("debug2 \n");

               /* frame # update */

               nframe++;
			   vhead.frame++;

           }
		   sec_count++;  //2012-12-03
		   vhead.sec_fm_ref++;


           // second data read end

           period++;
           if(runp.period > 0 && (runp.period == period)) goto exit_loop;

      }

exit_loop:

       printf(" ==========================================================\n");   

	   if(flag_fi_open==1) {
		   i=fclose(fi);
		   flag_fi_open=0;
	   }
	   if(flag_fo_open==1) {
		   i=fclose(fo);
		   flag_fo_open=0;
	   }

       printf("\n");
       printf("  VDIF file (%s) has been created.\n\n",runp.vdifname);



       free(obuf);

       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       printf("Time elapsed (sec) for One obs process is %f\n",telapsed);

exitend: 
	   if(flag_fi_open==1) i=fclose(fi);
	   if(flag_fo_open==1) i=fclose(fo);

       return 0;
}
